import logging
import time
from django.views.generic import View
from django.db.models import Q
from django.shortcuts import get_object_or_404, render
from django.contrib.auth.mixins import PermissionRequiredMixin
from django_tables2 import RequestConfig

from utilities.views import (
    GetReturnURLMixin, ObjectView, ObjectEditView, ObjectDeleteView, ObjectListView,
)
from utilities.paginator import EnhancedPaginator, get_paginate_count
from virtualization.models import VirtualMachine
from dcim.models import Device
from .models import VMWorkLog, DeviceWorkLog
from . import tables, filters, forms

# 
# Device Work Log
# 
class DeviceWorkLogListView(View):
    """
    Present a history of changes made to a particular virtual machine.
    """
    queryset = None
    filterset = filters.DeviceWorkLogFilterSet
    filterset_form = forms.DeviceWorkLogFilterForm
    table = tables.DeviceWorkLogTable
    template_name = 'netbox_work_logs/device/worklog/worklog_list.html'
    
    def get(self, request, device, **kwargs):
        logger = logging.getLogger('netbox_work_logs.views.DeviceWorkLogListView')
        # Handle QuerySet restriction of parent object if needed
        if hasattr(Device.objects, 'restrict'):
            obj = get_object_or_404(Device.objects.restrict(
                request.user, 'view'), id=device)
        else:
            obj = get_object_or_404(Device, id=device)
        
        self.queryset = DeviceWorkLog.objects.restrict(request.user, 'view').prefetch_related(
                'user', 'device'
            ).filter(
                Q(device=device)
            )
        if self.filterset:
            self.queryset = self.filterset(request.GET, self.queryset).qs

        table = tables.DeviceWorkLogTable(
            self.queryset,
            orderable=True
        )
        
        # Apply the request context
        paginate = {
            'paginator_class': EnhancedPaginator,
            'per_page': get_paginate_count(request)
        }
        RequestConfig(request, paginate).configure(table)

        context = {
            'table': table,
            'obj': obj,
            'filter_form': self.filterset_form(request.GET, label_suffix='') if self.filterset_form else None,
        }

        return render(request, self.template_name, context)
    
    def alter_queryset(self, request):
        # .all() is necessary to avoid caching queries
        return self.queryset.all()


class DeviceWorkLogDetailView(ObjectView):
    """
    Detail Device Work Log View
    """
    queryset = DeviceWorkLog.objects.prefetch_related('user', 'device')

    def get(self, request, pk):
        deviceworklog = get_object_or_404(self.queryset, pk=pk)
        return render(request, 'netbox_work_logs/device/worklog/worklog_detail.html', {'deviceworklog': deviceworklog})


class DeviceWorkLogEditView(ObjectEditView):
    queryset = DeviceWorkLog.objects.all()
    model_form = forms.DeviceWorkLogEditForm
    template_name = 'netbox_work_logs/device/worklog/worklog_edit.html'


class DeviceWorkLogDeleteView(ObjectDeleteView):
    queryset = DeviceWorkLog.objects.all()

# 
# VM Work Log
# 

class VMWorkLogListView(View):
    """
    Present a history of changes made to a particular virtual machine.
    """
    queryset = None
    filterset = filters.VMWorkLogFilterSet
    filterset_form = forms.VMWorkLogFilterForm
    table = tables.VMWorkLogTable
    template_name = 'netbox_work_logs/vm/worklog/worklog_list.html'
    
    def get(self, request, vm, **kwargs):
        logger = logging.getLogger('netbox_work_logs.views.VMWorkLogListView')
        # Handle QuerySet restriction of parent object if needed
        if hasattr(VirtualMachine.objects, 'restrict'):
            obj = get_object_or_404(VirtualMachine.objects.restrict(
                request.user, 'view'), id=vm)
        else:
            obj = get_object_or_404(VirtualMachine, id=vm)
            
        self.queryset = VMWorkLog.objects.restrict(request.user, 'view').prefetch_related(
                'user', 'vm'
            ).filter(
                Q(vm=vm)
            )

        if self.filterset:
            self.queryset = self.filterset(request.GET, self.queryset).qs

        # # Construct the table based on the user`s permissions
        # if request.user.is_authenticated:
        #     columns = request.user.config.get(f"tables.")
        # else:
        #     columns = None
            
        table = tables.VMWorkLogTable(
            self.queryset,
            orderable=True
        )
        
        # Apply the request context
        paginate = {
            'paginator_class': EnhancedPaginator,
            'per_page': get_paginate_count(request)
        }
        RequestConfig(request, paginate).configure(table)

        context = {
            'table': table,
            'obj': obj,
            'filter_form': self.filterset_form(request.GET, label_suffix='') if self.filterset_form else None,
        }
        
        return render(request, self.template_name, context)
    
    def alter_queryset(self, request):
        # .all() is necessary to avoid caching queries
        return self.queryset.all()
    
class VMWorkLogDetailView(ObjectView):
    """
    Detail VM work log view
    """
    queryset = VMWorkLog.objects.prefetch_related('user', 'vm')
    
    def get(self, request, pk):
        vmworklog = get_object_or_404(self.queryset, pk=pk)
        return render(request, 'netbox_work_logs/vm/worklog/worklog_detail.html', {'vmworklog': vmworklog})
    
class VMWorkLogEditView(ObjectEditView):
    queryset = VMWorkLog.objects.all()
    model_form = forms.VMWorkLogEditForm
    template_name = 'netbox_work_logs/vm/worklog/worklog_edit.html'

class VMWorkLogDeleteView(ObjectDeleteView):
    queryset = VMWorkLog.objects.all()


    



import django_filters
from django.contrib.auth.models import User
from django.db.models import Q

from utilities.filters import BaseFilterSet
from .models import Category, VMWorkLog, DeviceWorkLog

class DeviceWorkLogFilterSet(BaseFilterSet):
    """
    Device Work Log Filter Set
    """
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )
    time = django_filters.DateTimeFromToRangeFilter()
    user = django_filters.ModelMultipleChoiceFilter(
        field_name='user__id',
        queryset=User.objects.all(),
        to_field_name='id',
        label='Author',
    )
    category = django_filters.ModelMultipleChoiceFilter(
        field_name='category__id',
        queryset=Category.objects.all(),
        to_field_name='id',
        label='Category',
    )

    class Meta:
        model = DeviceWorkLog
        fields = [
            'id', 'category', 'subject', 'user', 'content', 'internal_only', 'log_id', 'ticket_id', 'time'
        ]
    
    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(category__category__icontains=value) |
            Q(user__username__icontains=value) |
            Q(content__icontains=value) |
            Q(ticket_id__icontains=value)
        )

    
class VMWorkLogFilterSet(BaseFilterSet):
    """
    Virtual Machine Work Log Filter Set
    """
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )
    time = django_filters.DateTimeFromToRangeFilter()
    user = django_filters.ModelMultipleChoiceFilter(
        field_name='user__id',
        queryset=User.objects.all(),
        to_field_name='id',
        label='Author',
    )
    category = django_filters.ModelMultipleChoiceFilter(
        field_name='category__id',
        queryset=Category.objects.all(),
        to_field_name='id',
        label='Category',
    )

    class Meta:
        model = VMWorkLog
        fields = [
            'id', 'category', 'subject', 'user', 'content', 'internal_only', 'log_id', 'ticket_id', 'time'
        ]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(category__category__icontains=value) |
            Q(user__username__icontains=value) |
            Q(content__icontains=value) |
            Q(ticket_id__icontains=value)
        )



from django.contrib.auth.context_processors import PermWrapper

from extras.plugins import PluginTemplateExtension
from .models import VMWorkLog, DeviceWorkLog, Category


class DeviceWorkLogs(PluginTemplateExtension):
    """
    Display work logs block on individual device page.
    """
    model = 'dcim.device'
    
    def buttons(self):
        context = {
            'perms': PermWrapper(self.context['request'].user),
        }
        return self.render('netbox_work_logs/device/inc/logslist_btn.html', context)
    
    def right_page(self):
        
        # get latest 5 logs
        logs = DeviceWorkLog.objects.filter(
            device=self.context['object'].id).order_by('-time')[:5]
        
        return self.render('netbox_work_logs/device/latest_logs.html', extra_context={
            'work_logs': logs,
        })
        
    
class VMWorkLogs(PluginTemplateExtension):
    """
    Display work logs block on individual virtual machine page.
    """
    model = 'virtualization.virtualmachine'
    
    def buttons(self):
        
        context = {
            'perms': PermWrapper(self.context['request'].user),
        }
       
        return self.render('netbox_work_logs/vm/inc/logslist_btn.html', context)
    
    def right_page(self):
        
        # get latest 5 logs
        logs = VMWorkLog.objects.filter(
            vm=self.context['object'].id).order_by('-time')[:5]
        
        return self.render('netbox_work_logs/vm/latest_logs.html', extra_context={
            'work_logs': logs,
        })

# PluginTemplateExtension subclasses must be package into an iterable named
# template_extensions to be imported by NetBox.
template_extensions = [DeviceWorkLogs, VMWorkLogs]



from django.urls import path
from . import views

urlpatterns = [
    path('vm-worklogs/<int:vm>/list/',
         views.VMWorkLogListView.as_view(), name='vmworklog_list'),
    path('vm-worklogs/<int:pk>/detail/',
         views.VMWorkLogDetailView.as_view(), name='vmworklog_detail'),
    path('vm-worklogs/<int:pk>/edit/',
         views.VMWorkLogEditView.as_view(), name='vmworklog_edit'),
    path('vm-worklogs/<int:pk>/delete/',
         views.VMWorkLogDeleteView.as_view(), name='vmworklog_delete'),
    path('device-worklogs/<int:device>/list/', 
         views.DeviceWorkLogListView.as_view(), name='deviceworklog_list'),
    path('device-worklogs/<int:pk>/detail/',
         views.DeviceWorkLogDetailView.as_view(), name='deviceworklog_detail'),
    path('device-worklogs/<int:pk>/edit/',
         views.DeviceWorkLogEditView.as_view(), name='deviceworklog_edit'),
    path('device-worklogs/<int:pk>/delete/',
         views.DeviceWorkLogDeleteView.as_view(), name='deviceworklog_delete'),
    
]



import uuid
from django.db import models
from django.contrib.auth.models import User
from django.urls import reverse

from utilities.querysets import RestrictedQuerySet


class Category(models.Model):
    """
    Work log category model
    """
    category=models.CharField(max_length=50)
    
    class Meta:
        managed = True
        verbose_name = 'Category'
        verbose_name_plural = 'Categories'
    
    def __str__(self):
        return self.category


class DeviceWorkLog(models.Model):
    """
    The Device work log entry is tracked in the Device work logs table.
    """
    time = models.DateTimeField(
        auto_now=True,
        editable=True,
        db_index=True
    )
    user = models.ForeignKey(
        to=User,
        on_delete=models.SET_NULL,
        related_name='devicelogs',
        blank=True,
        null=True
    )
    log_id = models.UUIDField(
        default=uuid.uuid4,
        editable=False
    )
    category = models.ForeignKey(
        to=Category,
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )
    subject = models.CharField(
        max_length=200,
        blank=True,
        null=True
    )
    content = models.TextField(
        blank=True,
        null=True
    )
    internal_only = models.BooleanField(
        default=False
    )
    ticket_id = models.CharField(
        max_length=50,
        blank=True,
        null=True
    )
    device = models.ForeignKey(
        to="dcim.Device", 
        on_delete=models.CASCADE
    )
    
    objects = RestrictedQuerySet.as_manager()

    class Meta:
        ordering = ["-time"]
    
    def __str__(self):
        return self.subject

    def get_absolute_url(self):
        return reverse("plugins:netbox_work_logs:deviceworklog_detail", kwargs={"pk": self.pk})
    
class VMWorkLog(models.Model):
    """
    The VM work log entry is tracked in the VM work logs table.
    """
    time = models.DateTimeField(
        auto_now=True,
        editable=False, 
        db_index=True
    )
    user = models.ForeignKey(
        to=User, 
        on_delete=models.SET_NULL,
        related_name='vmlogs',
        blank=True,
        null=True
    )
    log_id = models.UUIDField(
        default=uuid.uuid4, 
        editable=False
    )
    category = models.ForeignKey(
        to=Category, 
        on_delete=models.SET_NULL, 
        blank=True, 
        null=True
    )
    subject = models.CharField(
        max_length=200, 
        blank=True, 
        null=True
    )
    content = models.TextField(
        blank=True, 
        null=True
    )
    internal_only = models.BooleanField(
        default=False
    )
    ticket_id = models.CharField(
        max_length=50,
        blank=True,
        null=True
    )
    vm = models.ForeignKey(
        to="virtualization.VirtualMachine", 
        on_delete=models.CASCADE
    )
    
    objects = RestrictedQuerySet.as_manager()

    class Meta:
        ordering = ["-time"]
        
    def __str__(self):
        return self.subject
        
    def get_absolute_url(self):
        return reverse("plugins:netbox_work_logs:vmworklog_detail", kwargs={"pk": self.pk})
    


    



from extras.plugins import PluginConfig

class WorkLogsConfig(PluginConfig):
    """
    This class defines attributes for the NetBox Work Logs plugin.
    """
    # Plugin package name
    name = 'netbox_work_logs'

    # Human-friendly name and description
    verbose_name = 'Work Logs'
    description = 'A plugin to render and edit work logs for each virtual machine and each device'

    # Plugin version
    version = '1.1'

    # Plugin author
    author = 'Vasilatos Vitaliy'
    author_email = 'vasilatos80@gmail.com'

    # Configuration parameters that MUST be defined by the user (if any)
    required_settings = []

    # Default configuration parameter values, if not set by the user
    # default_settings = {
    #     'loud': True
    # }

    # Base URL path. If not set, the plugin name will be used.
    base_url = 'work-logs'

    # Caching config
    # caching_config = {}


config = WorkLogsConfig



from django.contrib import admin
from .models import Category

@admin.register(Category)
class CategoryAdmin(admin.ModelAdmin):
    """
    Define Work Log category
    """
    list_display=('category',)


import django_tables2 as tables

from utilities.tables import (
    BaseTable, BooleanColumn, ButtonsColumn, ColorColumn, ColoredLabelColumn, TagColumn, ToggleColumn,
)
from .models import VMWorkLog, DeviceWorkLog

WORKLOG_TIME = """
<a href="{{ record.get_absolute_url }}">{{ value|date:"SHORT_DATETIME_FORMAT" }}</a>
"""
WORKLOG_SUBJECT = """
{% if record.get_absolute_url %}
    <a href="{{ record.get_absolute_url }}">{{ record.subject }}</a>
{% else %}
    {{ record.subject }}
{% endif %}
"""
WORKLOG_CATEGORY = """
<span class="label label-success">{{ record.category.category }}</span>
"""

class WorkLogsButtonsColumn(tables.TemplateColumn):
    """
    Render edit and delete buttons for a Device or a VM work log
    """
    buttons = ('edit', 'delete')
    attrs = {'td': {'class': 'text-right text-nowrap noprint'}}
    
    # Note that braces are escaped to allow for string formatting prior to template rendering
    template_code = """
    {{% if "edit" in buttons and perms.{app_label}.change_{model_name} %}}
        <a href="{{% url 'plugins:{app_label}:{model_name}_edit' {pk_field}=record.{pk_field} %}}?return_url={{{{ request.path }}}}" class="btn btn-xs btn-warning" title="Edit">
            <i class="fa fa-pencil"></i>
        </a>
    {{% endif %}}
    {{% if "delete" in buttons and perms.{app_label}.delete_{model_name} %}}
        <a href="{{% url 'plugins:{app_label}:{model_name}_delete' {pk_field}=record.{pk_field} %}}?return_url={{{{ request.path }}}}" class="btn btn-xs btn-danger" title="Delete">
            <i class="fa fa-trash"></i>
        </a>
    {{% endif %}}
    """
    
    def __init__(self, model, *args, pk_field='pk', buttons=None, prepend_template=None, **kwargs):
        if prepend_template:
            prepend_template = prepend_template.replace('{', '{{')
            prepend_template = prepend_template.replace('}', '}}')
            self.template_code = prepend_template + self.template_code

        template_code = self.template_code.format(
            app_label=model._meta.app_label,
            model_name=model._meta.model_name,
            pk_field=pk_field,
            buttons=buttons
        )

        super().__init__(template_code=template_code, *args, **kwargs)

        self.extra_context.update({
            'buttons': buttons or self.buttons,
        })

    def header(self):
        return ''
    

class DeviceWorkLogTable(BaseTable):
    """
    Device Work Log Table
    """
    time = tables.TemplateColumn(
        template_code=WORKLOG_TIME,
        verbose_name='Time'
    )
    subject = tables.TemplateColumn(
        template_code=WORKLOG_SUBJECT,
        verbose_name='Subject'
    )
    category = tables.TemplateColumn(
        template_code=WORKLOG_CATEGORY,
        verbose_name='Category'
    )
    user = tables.Column(
        verbose_name='Author'
    )
    ticket_id = tables.Column(
        verbose_name='Ticket ID'
    )
    log_id = tables.Column(
        verbose_name='Log ID'
    )

    actions = WorkLogsButtonsColumn(DeviceWorkLog, pk_field='pk')

    class Meta(BaseTable.Meta):
        model = VMWorkLog
        fields = ('pk', 'time', 'subject', 'category',
                  'user', 'ticket_id', 'log_id', 'actions')
        default_columns = ('time', 'subject', 'category',
                           'user', 'ticket_id', 'log_id', 'actions')


class VMWorkLogTable(BaseTable):
    """
    Virtual Machine Work Log Table
    """
    time = tables.TemplateColumn(
        template_code=WORKLOG_TIME,
        verbose_name='Time'
    )
    subject = tables.TemplateColumn(
        template_code=WORKLOG_SUBJECT,
        verbose_name='Subject'
    )
    category = tables.TemplateColumn(
        template_code=WORKLOG_CATEGORY,
        verbose_name='Category'
    )
    user = tables.Column(
        verbose_name='Author'
    )
    ticket_id = tables.Column(
        verbose_name='Ticket ID'
    )
    log_id = tables.Column(
        verbose_name='Log ID'
    )

    actions = WorkLogsButtonsColumn(VMWorkLog, pk_field='pk')

    class Meta(BaseTable.Meta):
        model = VMWorkLog
        fields = ('pk', 'time', 'subject', 'category', 'user', 'ticket_id', 'log_id', 'actions')
        default_columns = ('time', 'subject', 'category', 'user', 'ticket_id', 'log_id', 'actions')
        



from django import forms
from django.contrib.auth.models import User

from utilities.forms import (
    BootstrapMixin, DateTimePicker, DynamicModelChoiceField, CommentField, DynamicModelMultipleChoiceField,
    SmallTextarea, StaticSelect2, APISelectMultiple,
    BOOLEAN_WITH_BLANK_CHOICES
)
from .models import Category, VMWorkLog, DeviceWorkLog

# 
# Device Work Log Filter Form and Edit Form
# 

class DeviceWorkLogFilterForm(BootstrapMixin, forms.Form):
    """
    Device Work Log Filter Form
    """
    model = DeviceWorkLog
    q = forms.CharField(
        required=False,
        label='Search'
    )
    time_after = forms.DateTimeField(
        label='After',
        required=False,
        widget=DateTimePicker()
    )
    time_before = forms.DateTimeField(
        label='Before',
        required=False,
        widget=DateTimePicker()
    )
    category = DynamicModelMultipleChoiceField(
        queryset=Category.objects.all(),
        required=False,
        display_field='category',
        label='Category',
        widget=APISelectMultiple(
            api_url='/api/plugins/work-logs/categories/',
        )
    )
    user = DynamicModelMultipleChoiceField(
        queryset=User.objects.all(),
        required=False,
        display_field='username',
        label='Author',
        widget=APISelectMultiple(
            api_url='/api/users/users/',
        )
    )

class DeviceWorkLogEditForm(BootstrapMixin, forms.ModelForm):
    """
    Device Work Log Edit Form
    """
    category = forms.ModelChoiceField(
        queryset=Category.objects.all()
    )
    content = CommentField(
        widget=SmallTextarea,
        label='Content',
        required=False
    )
    internal_only = forms.NullBooleanField(
        widget=StaticSelect2(
            choices=BOOLEAN_WITH_BLANK_CHOICES
        ),
        label='Internal Only',
        required=False
    )

    class Meta:
        model = DeviceWorkLog
        fields = [
            'subject', 'category', 'internal_only', 'content', 'ticket_id'
        ]
    
# 
# Virtual Machine Work Log Filter Form and Edit Form
#   

class VMWorkLogFilterForm(BootstrapMixin, forms.Form):
    """
    Virtual Machine Work Log Filter Form
    """
    model = VMWorkLog
    q = forms.CharField(
        required=False,
        label='Search'
    )
    time_after = forms.DateTimeField(
        label='After',
        required=False,
        widget=DateTimePicker()
    )
    time_before = forms.DateTimeField(
        label='Before',
        required=False,
        widget=DateTimePicker()
    )
    category = DynamicModelMultipleChoiceField(
        queryset=Category.objects.all(),
        required=False,
        display_field='category',
        label='Category',
        widget=APISelectMultiple(
            api_url='/api/plugins/work-logs/categories/',
        )
    )
    # action = forms.ChoiceField(
    #     choices=add_blank_choice(ObjectChangeActionChoices),
    #     required=False,
    #     widget=StaticSelect2()
    # )
    user = DynamicModelMultipleChoiceField(
        queryset=User.objects.all(),
        required=False,
        display_field='username',
        label='Author',
        widget=APISelectMultiple(
            api_url='/api/users/users/',
        )
    )


class VMWorkLogEditForm(BootstrapMixin, forms.ModelForm):
    category = forms.ModelChoiceField(
        queryset=Category.objects.all()
    )
    content = CommentField(
        widget = SmallTextarea,
        label='Content',
        required=False
    )
    internal_only = forms.NullBooleanField(
        widget=StaticSelect2(
            choices=BOOLEAN_WITH_BLANK_CHOICES
        ),
        label='Internal Only',
        required=False
    )
    class Meta:
        model = VMWorkLog
        fields = [
            'subject', 'category', 'internal_only', 'content', 'ticket_id'
        ]
    



import os
import shutil
from setuptools import find_packages, setup
rootdir = os.path.abspath(os.path.dirname(__file__))
os.makedirs('/opt/netbox/docs/models/netbox_work_logs', exist_ok=True)
shutil.copy2(rootdir + '/netbox_work_logs/doc/vmworklog.md', 
             '/opt/netbox/docs/models/netbox_work_logs')
shutil.copy2(rootdir + '/netbox_work_logs/doc/deviceworklog.md',
             '/opt/netbox/docs/models/netbox_work_logs')
setup(
    name='netbox-work-logs',
    version='1.1',
    description='A NetBox work logs plugin',
    url='https://github.com/vas-git/netbox-work-logs',
    author='Vasilatos Vitaliy',
    license='Apache 2.0',
    install_requires=[],
    packages=find_packages(),
    include_package_data=True,
)
